@@ -1,3 +1,8 @@
+from . import config
+from ._async import wrap_coroutine
+import asyncio
+import copyreg
+import dateutil
 import datetime
 import functools
 import sys
@@ -14,21 +19,40 @@ import inspect
 from dateutil import parser
 from dateutil.tz import tzlocal
 
-
 try:
     from maya import MayaDT
 except ImportError:
     MayaDT = None
 
+_TIME_NS_PRESENT = hasattr(time, 'time_ns')
+_MONOTONIC_NS_PRESENT = hasattr(time, 'monotonic_ns')
+_PERF_COUNTER_NS_PRESENT = hasattr(time, 'perf_counter_ns')
+_EPOCH = datetime.datetime(1970, 1, 1)
+_EPOCHTZ = datetime.datetime(1970, 1, 1, tzinfo=dateutil.tz.UTC)
 
 real_time = time.time
 real_localtime = time.localtime
 real_gmtime = time.gmtime
+real_monotonic = time.monotonic
+real_perf_counter = time.perf_counter
 real_strftime = time.strftime
 real_date = datetime.date
 real_datetime = datetime.datetime
-real_date_objects = [real_time, real_localtime, real_gmtime, real_strftime, real_date, real_datetime]
-_real_time_object_ids = set(id(obj) for obj in real_date_objects)
+real_date_objects = [real_time, real_localtime, real_gmtime, real_monotonic, real_perf_counter, real_strftime, real_date, real_datetime]
+
+if _TIME_NS_PRESENT:
+    real_time_ns = time.time_ns
+    real_date_objects.append(real_time_ns)
+
+if _MONOTONIC_NS_PRESENT:
+    real_monotonic_ns = time.monotonic_ns
+    real_date_objects.append(real_monotonic_ns)
+
+if _PERF_COUNTER_NS_PRESENT:
+    real_perf_counter_ns = time.perf_counter_ns
+    real_date_objects.append(real_perf_counter_ns)
+
+_real_time_object_ids = {id(obj) for obj in real_date_objects}
 
 # time.clock is deprecated and was removed in Python 3.8
 real_clock = getattr(time, 'clock', None)
@@ -38,13 +62,6 @@ tz_offsets = []
 ignore_lists = []
 tick_flags = []
 
-# Python3 doesn't have basestring, but it does have str.
-try:
-    # noinspection PyUnresolvedReferences
-    string_type = basestring
-except NameError:
-    string_type = str
-
 try:
     # noinspection PyUnresolvedReferences
     real_uuid_generate_time = uuid._uuid_generate_time
@@ -65,20 +82,6 @@ try:
 except (AttributeError, ImportError):
     real_uuid_create = None
 
-try:
-    import copy_reg as copyreg
-except ImportError:
-    import copyreg
-
-try:
-    iscoroutinefunction = inspect.iscoroutinefunction
-    from freezegun._async import wrap_coroutine
-except AttributeError:
-    iscoroutinefunction = lambda x: False
-
-    def wrap_coroutine(*args):
-        raise NotImplementedError()
-
 
 # keep a cache of module attributes otherwise freezegun will need to analyze too many modules all the time
 _GLOBAL_MODULES_CACHE = {}
@@ -115,7 +118,7 @@ def _get_module_attributes_hash(module):
         module_dir = dir(module)
     except (ImportError, TypeError):
         module_dir = []
-    return '{0}-{1}'.format(id(module), hash(frozenset(module_dir)))
+    return f'{id(module)}-{hash(frozenset(module_dir))}'
 
 
 def _get_cached_module_attributes(module):
@@ -130,11 +133,6 @@ def _get_cached_module_attributes(module):
     return cached_attrs
 
 
-# Stolen from six
-def with_metaclass(meta, *bases):
-    """Create a base class with a metaclass."""
-    return meta("NewBase", bases, {})
-
 _is_cpython = (
     hasattr(platform, 'python_implementation') and
     platform.python_implementation().lower() == "cpython"
@@ -148,6 +146,10 @@ def _should_use_real_time():
     if not call_stack_inspection_limit:
         return False
 
+    # Means stop() has already been called, so we can now return the real time
+    if not ignore_lists:
+        return True
+
     if not ignore_lists[-1]:
         return False
 
@@ -175,6 +177,12 @@ def fake_time():
     current_time = get_current_time()
     return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
 
+if _TIME_NS_PRESENT:
+    def fake_time_ns():
+        if _should_use_real_time():
+            return real_time_ns()
+        return int(int(fake_time()) * 1e9)
+
 
 def fake_localtime(t=None):
     if t is not None:
@@ -193,6 +201,53 @@ def fake_gmtime(t=None):
     return get_current_time().timetuple()
 
 
+def _get_fake_monotonic():
+    # For monotonic timers like .monotonic(), .perf_counter(), etc
+    current_time = get_current_time()
+    return (
+        calendar.timegm(current_time.timetuple()) +
+        current_time.microsecond / 1e6
+    )
+
+
+def _get_fake_monotonic_ns():
+    # For monotonic timers like .monotonic(), .perf_counter(), etc
+    current_time = get_current_time()
+    return (
+        calendar.timegm(current_time.timetuple()) * 1000000 +
+        current_time.microsecond
+    ) * 1000
+
+
+def fake_monotonic():
+    if _should_use_real_time():
+        return real_monotonic()
+
+    return _get_fake_monotonic()
+
+
+def fake_perf_counter():
+    if _should_use_real_time():
+        return real_perf_counter()
+
+    return _get_fake_monotonic()
+
+
+if _MONOTONIC_NS_PRESENT:
+    def fake_monotonic_ns():
+        if _should_use_real_time():
+            return real_monotonic_ns()
+
+        return _get_fake_monotonic_ns()
+
+
+if _PERF_COUNTER_NS_PRESENT:
+    def fake_perf_counter_ns():
+        if _should_use_real_time():
+            return real_perf_counter_ns()
+        return _get_fake_monotonic_ns()
+
+
 def fake_strftime(format, time_to_format=None):
     if time_to_format is None:
         if not _should_use_real_time():
@@ -250,10 +305,7 @@ def date_to_fakedate(date):
                     date.day)
 
 
-class FakeDate(with_metaclass(FakeDateMeta, real_date)):
-    def __new__(cls, *args, **kwargs):
-        return real_date.__new__(cls, *args, **kwargs)
-
+class FakeDate(real_date, metaclass=FakeDateMeta):
     def __add__(self, other):
         result = real_date.__add__(self, other)
         if result is NotImplemented:
@@ -296,10 +348,7 @@ class FakeDatetimeMeta(FakeDateMeta):
         return issubclass(subclass, real_datetime)
 
 
-class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
-    def __new__(cls, *args, **kwargs):
-        return real_datetime.__new__(cls, *args, **kwargs)
-
+class FakeDatetime(real_datetime, FakeDate, metaclass=FakeDatetimeMeta):
     def __add__(self, other):
         result = real_datetime.__add__(self, other)
         if result is NotImplemented:
@@ -320,6 +369,19 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
             tz = tzlocal()
         return datetime_to_fakedatetime(real_datetime.astimezone(self, tz))
 
+    @classmethod
+    def fromtimestamp(cls, t, tz=None):
+        if tz is None:
+            return real_datetime.fromtimestamp(
+                    t, tz=dateutil.tz.tzoffset("freezegun", cls._tz_offset())
+                ).replace(tzinfo=None)
+        return datetime_to_fakedatetime(real_datetime.fromtimestamp(t, tz))
+
+    def timestamp(self):
+        if self.tzinfo is None:
+            return (self - _EPOCH - self._tz_offset()).total_seconds()
+        return (self - _EPOCHTZ).total_seconds()
+
     @classmethod
     def now(cls, tz=None):
         now = cls._time_to_freeze() or real_datetime.now()
@@ -331,7 +393,7 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
 
     def date(self):
         return date_to_fakedate(self)
-    
+
     @property
     def nanosecond(self):
         try:
@@ -346,7 +408,7 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
 
     @classmethod
     def utcnow(cls):
-        result = cls._time_to_freeze() or real_datetime.utcnow()
+        result = cls._time_to_freeze() or real_datetime.now(datetime.timezone.utc)
         return datetime_to_fakedatetime(result)
 
     @staticmethod
@@ -401,14 +463,14 @@ def _parse_time_to_freeze(time_to_freeze_str):
     :returns: a naive ``datetime.datetime`` object
     """
     if time_to_freeze_str is None:
-        time_to_freeze_str = datetime.datetime.utcnow()
+        time_to_freeze_str = datetime.datetime.now(datetime.timezone.utc)
 
     if isinstance(time_to_freeze_str, datetime.datetime):
         time_to_freeze = time_to_freeze_str
     elif isinstance(time_to_freeze_str, datetime.date):
         time_to_freeze = datetime.datetime.combine(time_to_freeze_str, datetime.time())
     elif isinstance(time_to_freeze_str, datetime.timedelta):
-        time_to_freeze = datetime.datetime.utcnow() + time_to_freeze_str
+        time_to_freeze = datetime.datetime.now(datetime.timezone.utc) + time_to_freeze_str
     else:
         time_to_freeze = parser.parse(time_to_freeze_str)
 
@@ -422,7 +484,7 @@ def _parse_tz_offset(tz_offset):
         return datetime.timedelta(hours=tz_offset)
 
 
-class TickingDateTimeFactory(object):
+class TickingDateTimeFactory:
 
     def __init__(self, time_to_freeze, start):
         self.time_to_freeze = time_to_freeze
@@ -432,7 +494,7 @@ class TickingDateTimeFactory(object):
         return self.time_to_freeze + (real_datetime.now() - self.start)
 
 
-class FrozenDateTimeFactory(object):
+class FrozenDateTimeFactory:
 
     def __init__(self, time_to_freeze):
         self.time_to_freeze = time_to_freeze
@@ -454,7 +516,7 @@ class FrozenDateTimeFactory(object):
         self.tick(delta=delta)
 
 
-class StepTickTimeFactory(object):
+class StepTickTimeFactory:
 
     def __init__(self, time_to_freeze, step_width):
         self.time_to_freeze = time_to_freeze
@@ -480,10 +542,9 @@ class StepTickTimeFactory(object):
         self.tick(delta=delta)
 
 
-class _freeze_time(object):
+class _freeze_time:
 
-
-    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg, auto_tick_seconds):
+    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds):
         self.time_to_freeze = _parse_time_to_freeze(time_to_freeze_str)
         self.tz_offset = _parse_tz_offset(tz_offset)
         self.ignore = tuple(ignore)
@@ -492,22 +553,24 @@ class _freeze_time(object):
         self.undo_changes = []
         self.modules_at_start = set()
         self.as_arg = as_arg
+        self.as_kwarg = as_kwarg
 
     def __call__(self, func):
         if inspect.isclass(func):
             return self.decorate_class(func)
-        elif iscoroutinefunction(func):
+        elif inspect.iscoroutinefunction(func):
             return self.decorate_coroutine(func)
         return self.decorate_callable(func)
 
     def decorate_class(self, klass):
         if issubclass(klass, unittest.TestCase):
-            # If it's a TestCase, we assume you want to freeze the time for the
-            # tests, from setUpClass to tearDownClass
+            # If it's a TestCase, we freeze time around setup and teardown, as well
+            # as for every test case. This requires some care to avoid freezing
+            # the time pytest sees, as otherwise this would distort the reported
+            # timings.
 
-            # Use getattr as in Python 2.6 they are optional
-            orig_setUpClass = getattr(klass, 'setUpClass', None)
-            orig_tearDownClass = getattr(klass, 'tearDownClass', None)
+            orig_setUpClass = klass.setUpClass
+            orig_tearDownClass = klass.tearDownClass
 
             # noinspection PyDecorator
             @classmethod
@@ -515,10 +578,12 @@ class _freeze_time(object):
                 self.start()
                 if orig_setUpClass is not None:
                     orig_setUpClass()
+                self.stop()
 
             # noinspection PyDecorator
             @classmethod
             def tearDownClass(cls):
+                self.start()
                 if orig_tearDownClass is not None:
                     orig_tearDownClass()
                 self.stop()
@@ -526,20 +591,36 @@ class _freeze_time(object):
             klass.setUpClass = setUpClass
             klass.tearDownClass = tearDownClass
 
+            orig_setUp = klass.setUp
+            orig_tearDown = klass.tearDown
+
+            def setUp(*args, **kwargs):
+                self.start()
+                if orig_setUp is not None:
+                    orig_setUp(*args, **kwargs)
+
+            def tearDown(*args, **kwargs):
+                if orig_tearDown is not None:
+                    orig_tearDown(*args, **kwargs)
+                self.stop()
+
+            klass.setUp = setUp
+            klass.tearDown = tearDown
+
             return klass
 
         else:
 
             seen = set()
 
-            klasses = klass.mro() if hasattr(klass, 'mro') else [klass] + list(klass.__bases__)
+            klasses = klass.mro()
             for base_klass in klasses:
                 for (attr, attr_value) in base_klass.__dict__.items():
                     if attr.startswith('_') or attr in seen:
                         continue
                     seen.add(attr)
 
-                    if not callable(attr_value) or inspect.isclass(attr_value):
+                    if not callable(attr_value) or inspect.isclass(attr_value) or isinstance(attr_value, staticmethod):
                         continue
 
                     try:
@@ -578,6 +659,8 @@ class _freeze_time(object):
         datetime.date = FakeDate
 
         time.time = fake_time
+        time.monotonic = fake_monotonic
+        time.perf_counter = fake_perf_counter
         time.localtime = fake_localtime
         time.gmtime = fake_gmtime
         time.strftime = fake_strftime
@@ -595,18 +678,32 @@ class _freeze_time(object):
             ('real_datetime', real_datetime, FakeDatetime),
             ('real_gmtime', real_gmtime, fake_gmtime),
             ('real_localtime', real_localtime, fake_localtime),
+            ('real_monotonic', real_monotonic, fake_monotonic),
+            ('real_perf_counter', real_perf_counter, fake_perf_counter),
             ('real_strftime', real_strftime, fake_strftime),
             ('real_time', real_time, fake_time),
         ]
 
+        if _TIME_NS_PRESENT:
+            time.time_ns = fake_time_ns
+            to_patch.append(('real_time_ns', real_time_ns, fake_time_ns))
+
+        if _MONOTONIC_NS_PRESENT:
+            time.monotonic_ns = fake_monotonic_ns
+            to_patch.append(('real_monotonic_ns', real_monotonic_ns, fake_monotonic_ns))
+
+        if _PERF_COUNTER_NS_PRESENT:
+            time.perf_counter_ns = fake_perf_counter_ns
+            to_patch.append(('real_perf_counter_ns', real_perf_counter_ns, fake_perf_counter_ns))
+
         if real_clock is not None:
             # time.clock is deprecated and was removed in Python 3.8
             time.clock = fake_clock
             to_patch.append(('real_clock', real_clock, fake_clock))
 
         self.fake_names = tuple(fake.__name__ for real_name, real, fake in to_patch)
-        self.reals = dict((id(fake), real) for real_name, real, fake in to_patch)
-        fakes = dict((id(real), fake) for real_name, real, fake in to_patch)
+        self.reals = {id(fake): real for real_name, real, fake in to_patch}
+        fakes = {id(real): fake for real_name, real, fake in to_patch}
         add_change = self.undo_changes.append
 
         # Save the current loaded modules
@@ -630,6 +727,21 @@ class _freeze_time(object):
                         setattr(module, attribute_name, fake)
                         add_change((module, attribute_name, attribute_value))
 
+        # To avoid breaking `asyncio.sleep()`, let asyncio event loops see real
+        # monotonic time even though we've just frozen `time.monotonic()` which
+        # is normally used there. If we didn't do this, `await asyncio.sleep()`
+        # would be hanging forever breaking many tests that use `freeze_time`.
+        #
+        # Note that we cannot statically tell the class of asyncio event loops
+        # because it is not officially documented and can actually be changed
+        # at run time using `asyncio.set_event_loop_policy`. That's why we check
+        # the type by creating a loop here and destroying it immediately.
+        event_loop = asyncio.new_event_loop()
+        event_loop.close()
+        EventLoopClass = type(event_loop)
+        add_change((EventLoopClass, "time", EventLoopClass.time))
+        EventLoopClass.time = lambda self: real_monotonic()
+
         return freeze_factory
 
     def stop(self):
@@ -643,8 +755,8 @@ class _freeze_time(object):
             datetime.date = real_date
             copyreg.dispatch_table.pop(real_datetime)
             copyreg.dispatch_table.pop(real_date)
-            for module, module_attribute, original_value in self.undo_changes:
-                setattr(module, module_attribute, original_value)
+            for module_or_object, attribute, original_value in self.undo_changes:
+                setattr(module_or_object, attribute, original_value)
             self.undo_changes = []
 
             # Restore modules loaded after start()
@@ -658,7 +770,7 @@ class _freeze_time(object):
                         continue
                     elif mod_name.startswith(self.ignore) or mod_name.endswith('.six.moves'):
                         continue
-                    elif (not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time')):
+                    elif not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time'):
                         continue
                     for module_attribute in dir(module):
 
@@ -675,11 +787,22 @@ class _freeze_time(object):
                             setattr(module, module_attribute, real)
 
             time.time = real_time
+            time.monotonic = real_monotonic
+            time.perf_counter = real_perf_counter
             time.gmtime = real_gmtime
             time.localtime = real_localtime
             time.strftime = real_strftime
             time.clock = real_clock
 
+            if _TIME_NS_PRESENT:
+                time.time_ns = real_time_ns
+
+            if _MONOTONIC_NS_PRESENT:
+                time.monotonic_ns = real_monotonic_ns
+
+            if _PERF_COUNTER_NS_PRESENT:
+                time.perf_counter_ns = real_perf_counter_ns
+
             if uuid_generate_time_attr:
                 setattr(uuid, uuid_generate_time_attr, real_uuid_generate_time)
             uuid._UuidCreate = real_uuid_create
@@ -691,22 +814,24 @@ class _freeze_time(object):
     def decorate_callable(self, func):
         def wrapper(*args, **kwargs):
             with self as time_factory:
-                if self.as_arg:
+                if self.as_arg and self.as_kwarg:
+                    assert False, "You can't specify both as_arg and as_kwarg at the same time. Pick one."
+                elif self.as_arg:
                     result = func(time_factory, *args, **kwargs)
+                elif self.as_kwarg:
+                    kwargs[self.as_kwarg] = time_factory
+                    result = func(*args, **kwargs)
                 else:
                     result = func(*args, **kwargs)
             return result
         functools.update_wrapper(wrapper, func)
 
-        # update_wrapper already sets __wrapped__ in Python 3.2+, this is only
-        # needed for Python 2.x support
-        wrapper.__wrapped__ = func
-
         return wrapper
 
 
-def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False, auto_tick_seconds=0):
-    acceptable_times = (type(None), string_type, datetime.date, datetime.timedelta,
+def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False, as_kwarg='',
+                auto_tick_seconds=0):
+    acceptable_times = (type(None), str, datetime.date, datetime.timedelta,
              types.FunctionType, types.GeneratorType)
 
     if MayaDT is not None:
@@ -715,34 +840,35 @@ def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_ar
     if not isinstance(time_to_freeze, acceptable_times):
         raise TypeError(('freeze_time() expected None, a string, date instance, datetime '
                          'instance, MayaDT, timedelta instance, function or a generator, but got '
-                         'type {0}.').format(type(time_to_freeze)))
+                         'type {}.').format(type(time_to_freeze)))
     if tick and not _is_cpython:
         raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')
 
     if isinstance(time_to_freeze, types.FunctionType):
-        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, auto_tick_seconds)
+        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds)
 
     if isinstance(time_to_freeze, types.GeneratorType):
-        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, auto_tick_seconds)
+        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds)
 
     if MayaDT is not None and isinstance(time_to_freeze, MayaDT):
         return freeze_time(time_to_freeze.datetime(), tz_offset, ignore,
-                           tick, as_arg)
+                           tick, as_arg, as_kwarg, auto_tick_seconds)
 
     if ignore is None:
         ignore = []
     ignore = ignore[:]
-    ignore.append('nose.plugins')
-    ignore.append('six.moves')
-    ignore.append('django.utils.six.moves')
-    ignore.append('google.gax')
-    ignore.append('threading')
-    ignore.append('Queue')
-    ignore.append('selenium')
-    ignore.append('_pytest.terminal.')
-    ignore.append('_pytest.runner.')
-    
-    return _freeze_time(time_to_freeze, tz_offset, ignore, tick, as_arg, auto_tick_seconds)
+    if config.settings.default_ignore_list:
+        ignore.extend(config.settings.default_ignore_list)
+
+    return _freeze_time(
+        time_to_freeze_str=time_to_freeze,
+        tz_offset=tz_offset,
+        ignore=ignore,
+        tick=tick,
+        as_arg=as_arg,
+        as_kwarg=as_kwarg,
+        auto_tick_seconds=auto_tick_seconds,
+    )
 
 
 # Setup adapters for sqlite
